# Generated by Django 3.0.2 on 2020-04-21 14:36

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='BooleanField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('required', models.BooleanField(blank=True, default=True)),
                ('order', models.IntegerField(blank=True, default=0)),
                ('help', models.TextField(blank=True, default='')),
                ('config', models.TextField(blank=True, default='{}')),
                ('default', models.BooleanField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('description', models.TextField(blank=True, default='')),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='children', to='drf_schemas.Category')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Choice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='ChoicesField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('required', models.BooleanField(blank=True, default=True)),
                ('order', models.IntegerField(blank=True, default=0)),
                ('help', models.TextField(blank=True, default='')),
                ('config', models.TextField(blank=True, default='{}')),
                ('multi', models.BooleanField(blank=True, default=False)),
                ('default', models.ManyToManyField(related_name='field_default', to='drf_schemas.Choice')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='DateTimeField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('required', models.BooleanField(blank=True, default=True)),
                ('order', models.IntegerField(blank=True, default=0)),
                ('help', models.TextField(blank=True, default='')),
                ('config', models.TextField(blank=True, default='{}')),
                ('default', models.DateTimeField(blank=True, null=True)),
                ('min_value', models.DateTimeField(blank=True, null=True)),
                ('max_value', models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='File',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('file', models.FileField(upload_to='')),
                ('size_bytes', models.IntegerField(blank=True, default=0)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='FileField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('required', models.BooleanField(blank=True, default=True)),
                ('order', models.IntegerField(blank=True, default=0)),
                ('help', models.TextField(blank=True, default='')),
                ('config', models.TextField(blank=True, default='{}')),
                ('multi', models.BooleanField(blank=True, default=False)),
                ('default', models.ManyToManyField(related_name='defaults', to='drf_schemas.File')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('image', models.ImageField(upload_to='')),
                ('size_bytes', models.IntegerField(blank=True, default=0)),
                ('height', models.IntegerField(blank=True, default=0)),
                ('width', models.IntegerField(blank=True, default=0)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ImagesField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('required', models.BooleanField(blank=True, default=True)),
                ('order', models.IntegerField(blank=True, default=0)),
                ('help', models.TextField(blank=True, default='')),
                ('config', models.TextField(blank=True, default='{}')),
                ('multi', models.BooleanField(blank=True, default=False)),
                ('default', models.ManyToManyField(related_name='defaults', to='drf_schemas.Image')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Item',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ItemField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('required', models.BooleanField(blank=True, default=True)),
                ('order', models.IntegerField(blank=True, default=0)),
                ('help', models.TextField(blank=True, default='')),
                ('config', models.TextField(blank=True, default='{}')),
                ('multi', models.BooleanField(blank=True, default=False)),
                ('default', models.ManyToManyField(related_name='field_default', to='drf_schemas.Item')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ItemSchema',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('config', models.TextField(blank=True, default='{}')),
                ('category', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='schemas', to='drf_schemas.Category')),
                ('image', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='schema_images', to='drf_schemas.Image')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='NumberField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('required', models.BooleanField(blank=True, default=True)),
                ('order', models.IntegerField(blank=True, default=0)),
                ('help', models.TextField(blank=True, default='')),
                ('config', models.TextField(blank=True, default='{}')),
                ('default', models.FloatField(blank=True, null=True)),
                ('min_value', models.FloatField(blank=True, null=True)),
                ('max_value', models.FloatField(blank=True, null=True)),
                ('item_schema', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='number_fields', to='drf_schemas.ItemSchema')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TextField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('required', models.BooleanField(blank=True, default=True)),
                ('order', models.IntegerField(blank=True, default=0)),
                ('help', models.TextField(blank=True, default='')),
                ('config', models.TextField(blank=True, default='{}')),
                ('default', models.TextField(blank=True, default='')),
                ('item_schema', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='text_fields', to='drf_schemas.ItemSchema')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TextValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('value', models.TextField(blank=True, default='')),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='text_values', to='drf_schemas.TextField')),
                ('item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='text_values', to='drf_schemas.Item')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='NumberValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('value', models.FloatField(blank=True, null=True)),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='number_values', to='drf_schemas.NumberField')),
                ('item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='number_values', to='drf_schemas.Item')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ItemValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='values', to='drf_schemas.ItemField')),
                ('item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='item_values', to='drf_schemas.Item')),
                ('value', models.ManyToManyField(related_name='target_values', to='drf_schemas.Item')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='itemfield',
            name='item_schema',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='item_fields', to='drf_schemas.ItemSchema'),
        ),
        migrations.AddField(
            model_name='itemfield',
            name='target_schema',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='target_fields', to='drf_schemas.ItemSchema'),
        ),
        migrations.AddField(
            model_name='item',
            name='schema',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items', to='drf_schemas.ItemSchema'),
        ),
        migrations.CreateModel(
            name='ImagesValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='images_values', to='drf_schemas.ImagesField')),
                ('item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='images_values', to='drf_schemas.Item')),
                ('value', models.ManyToManyField(related_name='images_values', to='drf_schemas.Image')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='imagesfield',
            name='item_schema',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='images_fields', to='drf_schemas.ItemSchema'),
        ),
        migrations.CreateModel(
            name='FileValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='file_values', to='drf_schemas.FileField')),
                ('item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='file_values', to='drf_schemas.Item')),
                ('value', models.ManyToManyField(related_name='file_values', to='drf_schemas.File')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='filefield',
            name='item_schema',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='file_fields', to='drf_schemas.ItemSchema'),
        ),
        migrations.CreateModel(
            name='DateTimeValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('value', models.DateTimeField(blank=True, null=True)),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='datetime_values', to='drf_schemas.DateTimeField')),
                ('item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='datetime_values', to='drf_schemas.Item')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='datetimefield',
            name='item_schema',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='datetime_fields', to='drf_schemas.ItemSchema'),
        ),
        migrations.CreateModel(
            name='ChoicesValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='choices_values', to='drf_schemas.ChoicesField')),
                ('item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='choices_values', to='drf_schemas.Item')),
                ('value', models.ManyToManyField(related_name='field_values', to='drf_schemas.Choice')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='choicesfield',
            name='item_schema',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='choices_fields', to='drf_schemas.ItemSchema'),
        ),
        migrations.AddField(
            model_name='choice',
            name='field',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='choices', to='drf_schemas.ChoicesField'),
        ),
        migrations.CreateModel(
            name='BooleanValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('value', models.BooleanField(blank=True, null=True)),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='boolean_values', to='drf_schemas.BooleanField')),
                ('item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='boolean_values', to='drf_schemas.Item')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='booleanfield',
            name='item_schema',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='boolean_fields', to='drf_schemas.ItemSchema'),
        ),
    ]



from threading import Thread

from django.db.models import QuerySet

from . import models
from . import serializers
from . import resources
import tablib
from import_export.resources import modelresource_factory

_field_params = [{
    'fields': 'boolean_fields',
    'values': 'boolean_values',
    'value_model': models.BooleanValue,
    'field_model': models.BooleanField,
    'field_serializer': serializers.BooleanFieldSerializer,
    'value_serializer': serializers.BooleanValueSerializer,
}, {
    'fields': 'choices_fields',
    'values': 'choices_values',
    'value_model': models.ChoicesValue,
    'field_model': models.ChoicesField,
    'field_serializer': serializers.ChoicesFieldSerializer,
    'value_serializer': serializers.ChoicesValueSerializer,
}, {
    'fields': 'datetime_fields',
    'values': 'datetime_values',
    'value_model': models.DateTimeValue,
    'field_model': models.DateTimeField,
    'field_serializer': serializers.DateTimeFieldSerializer,
    'value_serializer': serializers.DateTimeValueSerializer,
}, {
    'fields': 'number_fields',
    'values': 'number_values',
    'value_model': models.NumberValue,
    'field_model': models.NumberField,
    'field_serializer': serializers.NumberFieldSerializer,
    'value_serializer': serializers.NumberValueSerializer,
}, {
    'fields': 'images_fields',
    'values': 'images_values',
    'value_model': models.ImagesValue,
    'field_model': models.ImagesField,
    'field_serializer': serializers.ImagesFieldSerializer,
    'value_serializer': serializers.ImagesValueSerializer,
}, {
    'fields': 'file_fields',
    'values': 'file_values',
    'value_model': models.FileValue,
    'field_model': models.FileField,
    'field_serializer': serializers.FileFieldSerializer,
    'value_serializer': serializers.FileValueSerializer,
}, {
    'fields': 'text_fields',
    'values': 'text_values',
    'value_model': models.TextValue,
    'field_model': models.TextField,
    'field_serializer': serializers.TextFieldSerializer,
    'value_serializer': serializers.TextValueSerializer,
}, {
    'fields': 'item_fields',
    'values': 'item_values',
    'value_model': models.ItemValue,
    'field_model': models.ItemField,
    'field_serializer': serializers.ItemFieldSerializer,
    'value_serializer': serializers.ItemValueSerializer,
}]

_multi_values = (
    'choices_values',
    'item_values',
    'images_values',
    'file_values'
)


def update_item(item=None):

    if item is None:
        return

    schema = item.schema

    for param in _field_params:
        fields_queryset = getattr(schema, param['fields']).all()
        values_queryset = getattr(item, param['values']).all()

        if len(fields_queryset) or len(values_queryset):
            fields = {field.pk: field for field in fields_queryset}
            values = {value.field.pk: value for value in values_queryset}

            fields_pk = set(fields)
            values_pk = set(values)

            create_pk = fields_pk - values_pk
            delete_pk = values_pk - fields_pk

            value_model = param['value_model']

            for pk in create_pk:
                field = fields[pk]
                if param['values'] in _multi_values:
                    value = value_model.objects.create(
                        item=item,
                        field=field
                    )
                    default = field.default.all()
                    if len(default):
                        value.value.set(default)
                else:
                    value_model.objects.create(
                        item=item,
                        field=field,
                        value=field.default
                    )

            for pk in delete_pk:
                values[pk].delete()


class ItemUpdater(Thread):

    def __init__(self, queryset: QuerySet, daemon: bool = True):
        super().__init__(daemon=daemon)
        self.queryset = queryset

    def run(self):
        for item in self.queryset.iterator():
            update_item(item)


_ie_resources = [{
    'name': 'categories',
    'resource': resources.CategoryResource,
    'model': models.Category
}, {
    'name': 'choices',
    'resource': resources.ChoiceResource,
    'model': models.Choice
}, {
    'name': 'item_schemas',
    'resource': resources.ItemSchemaResource,
    'model': models.ItemSchema
}, {
    'name': 'boolean_fields',
    'resource': resources.BooleanFieldResource,
    'model': models.BooleanField
}, {
    'name': 'choices_fields',
    'resource': resources.ChoicesFieldResource,
    'model': models.ChoicesField
}, {
    'name': 'datetime_fields',
    'resource': resources.DateTimeFieldResource,
    'model': models.DateTimeField
}, {
    'name': 'file_fields',
    'resource': resources.FileFieldResource,
    'model': models.FileField
}, {
    'name': 'images_fields',
    'resource': resources.ImagesFieldResource,
    'model': models.ImagesField
}, {
    'name': 'item_fields',
    'resource': resources.ItemFieldResource,
    'model': models.ItemField
}, {
    'name': 'number_fields',
    'resource': resources.NumberFieldResource,
    'model': models.NumberField
}, {
    'name': 'text_fields',
    'resource': resources.TextFieldResources,
    'model': models.TextField
}]


def export_schemas(output: str = 'dict'):
    data = {}
    for param in _ie_resources:
        dataset = param['resource']().export()
        data[param['name']] = getattr(dataset, output)

    return data


def import_schemas(data: dict):
    for param in _ie_resources:
        model_resource = modelresource_factory(model=param['model'])()
        dataset = tablib.Dataset()
        dataset.dict = data[param['name']]
        result = model_resource.import_data(dataset, dry_run=True)
        if result.has_errors():
            return {param['name']: result.base_errors}

        model_resource.import_data(dataset, dry_run=False)


# def export_schemas(context: [dict, None] = None):
#     data = {}
#     for param in _schema_params:
#         serializer = param['serializer'](
#             param['model'].objects.all(),
#             context=context,
#             many=True
#         )
#         data[param['name']] = serializer.data
#
#     for name, values in _init_fields.items():
#         for field, value in values.items():
#             for item in data[name]:
#                 item[field] = value
#
#     return data
